# OBJ.BOTS.IRC - an IRC bot
#
# obj/bots/irc.py (internet relay chat)
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 19-09-2018 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" irc bot. """

__version__ = 18
__txt__ = "IRC bot using the OBJ framework"

## imports

from obj.base import Dotted, Selector, locked
from obj.bots.base import Bot
from obj.bots.dcc import DCC
from obj.bots.main import cfg, defaults, fleet, opts, users
from obj.event import Event
from obj.task import launch

import _thread
import logging
import os
import queue
import random
import re
import socket
import ssl
import textwrap
import time
import threading

## defines

defaults = Dotted(defaults, _default="")
defaults.blocking = 1
defaults.daemon = False
defaults.ipv6 = False
defaults.port = 6667
defaults.ssl = False

## init

def start():
    bot = IRC()
    bot.walk("obj.bots.cmds")
    bot.walk("obj.bots.irc")
    bot.start()
    return bot

## classes

class Cfg(Dotted):

    """ IRC config class. """
    pass

class Event(Event):

    """ IRC event class. """

    pass

class TextWrap(textwrap.TextWrapper):

    """ used to wrap text to 500 chars. """
    
    def __init__(self):
        super().__init__()
        self.drop_whitespace = False
        self.fix_sentence_endings = True
        self.replace_whitespace = True
        self.tabsize = 4
        self.width = 500

class IRC(Bot):

    """ The IRC bot. """

    cc = "!"

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.register("352", self.handler)
        self.register("366", self.handler)
        self.register("376", self.handle_connect)
        self.register("433", self.handler)
        self.register("ERROR", self.handle_error)
        self.register("MODE", self.handler)
        self.register("PING", self.handler)
        self.register("PONG", self.handler)
        self.register("PRIVMSG", self.handle_privmsg)
        self._connected = threading.Event()
        self._error = ""
        self._last = 0
        self._outqueue = queue.Queue()
        self._ready = threading.Event()
        self.buffer = []
        self.cfg = Cfg(defaults)
        if args:
            self.cfg.update(args[0])
        if kwargs:
            self.cfg.update(kwargs)
        self.channels = []
        self.lastline = ""
        self.joined = []
        self.nrsend = 0
        self.nrconnect = 0
        self.threaded = False

    def announce(self, txt):
        """ announce text on all channels. """
        for channel in self.channels:
            self.say(channel, txt)

    def close(self):
        """ close the socket. """
        self.sock.shutdown(1)
        self.sock.close()
        self.fsock.close()

    def connect(self):
        """ connect to self.cfg.server on self.cfg.port. """
        oldsock = None
        logging.warn("connect %s %s" % (self.cfg.server, self.cfg.port))
        if self.cfg.ipv6:
            oldsock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        else:
            oldsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        oldsock.setblocking(int(self.cfg.blocking))
        oldsock.settimeout(60.0)
        oldsock.connect((self.cfg.server, int(self.cfg.port or 6667)))
        oldsock.setblocking(int(self.cfg.blocking))
        oldsock.settimeout(700.0)
        if self.cfg.ssl:
            self.sock = ssl.wrap_socket(oldsock)
        else:
            self.sock = oldsock
        self.fsock = self.sock.makefile("r")
        os.set_inheritable(self.sock.fileno(), os.O_RDWR)
        return True

    def command(self, cmd, *args, direct=False):
        """ output a command to the IRC server. Use the direct flag to by pass outqueue. """
        if not args:
            self.raw(cmd)
            return
        if len(args) == 1:
            self.raw("%s %s" % (cmd.upper(), args[0]), direct)
            return
        if len(args) == 2:
            self.raw("%s %s :%s" % (cmd.upper(), args[0], " ".join(args[1:])), direct)
            return
        if len(args) == 3:
            self.raw("%s %s %s :%s" % (cmd.upper(), args[0], args[1], " ".join(args[2:])), direct)
            return

    def dispatch(self, event, loader=None):
        """ dispatch a function based on event.command instead of event.cmd. """
        if event.command:
            func = self.get_handler(event.command)
            if func:
                func(event)
                return
        super().dispatch(event)

    def get_event(self, txt=""):
        """ return an event from the IRC socket. """ 
        if not txt and not self.buffer:
            try:
                self.some()
            except socket.timeout as ex:
                e = Event()
                e.command = "ERROR"
                e.txt = str(ex)
                return e
        line = txt or self.buffer.pop(0)
        return self.parsing(line.rstrip())

    def handler(self, event):
        """ default handler. """
        cmd = event.command
        if cmd == "001":
            self.logon(self.cfg.username or self.cfg.nick ,self.cfg.realname or self.cfg.nick)
        elif cmd == "PING":
            self.pongcheck = True
            self.command("PONG", event.txt, direct=True)
        elif cmd == "PONG":
            self.pongcheck = False
        elif cmd == "352":
            args = event.arguments
            users.userhosts[args[5]] = args[2] + "@" + args[3]
        elif cmd == "366":
            self.joined.append(event.channel)
        elif cmd == "433":
            nick = event.target + "_"
            self.cfg.nick = nick
            self.command("NICK", nick)

    def handle_connect(self, event):
        """ function that is called upon connect. """
        if "servermodes" in self:
            self.raw("MODE %s %s" % (self.cfg.nick, self.cfg.servermodes), True)
        logging.warn("connected to %s" % self.cfg.server)
        self.joinall()
        self._connected.set()

    def handle_error(self, event):
        """ error handler. """
        logging.error(event.txt)
        self._error = event
        self.stop()
        self.start()

    def handle_privmsg(self, event):
        """ privmsg handler. """
        self.callback(event)
        users.userhosts[event.nick] = event.origin
        if event.txt.startswith("\001DCC"):
            try:
                dcc = DCC()
                dcc.walk("obj.bots.cmds")
                dcc.walk("obj.bots.irc")
                dcc.encoding = "utf-8"
                launch(dcc.connect, event)
                return
            except ConnectionRefusedError:
                return
        elif event.txt.startswith("\001VERSION"):
            self.command("NOTICE", event.channel, "\001VERSION OBJ.BOTS.IRC #%s - Framework to program bots - https://bitbucket.org/bthate/obj.bots.irc\001" % __version__)
            return
        if event.txt and event.txt[0] != self.cc:
            return
        txt = event.txt[1:]
        splt = txt.split()
        splt[0] = splt[0].lower()
        if not self.get_handler(splt[0]):
            return
        if not users.allowed(event.origin, "USER"):
            return
        e = Event(event)
        e.txt=" ".join(splt)
        e.parse()
        #self.put(e)
        super().dispatch(e)
 
    def joinall(self):
        """ join all channels registered in self.channels. """
        for channel in self.channels:
            self.command("JOIN", channel, direct=True)

    def logon(self, username, realname):
        """ logon to the IRC network. """
        self.raw("NICK %s" % self.cfg.nick, direct=True)
        self.raw("USER %s %s %s :%s" % (username, self.cfg.server, self.cfg.server, realname), direct=True)

    def loop(self):
        """ basic event loop that reconnects on connection reset. """
        try:
            super().loop()
        except ConnectionResetError:
            self.start()

    def output(self):
        """ output loop. """
        while not self._stopped:
            channel, txt = self._outqueue.get()
            if not channel and not txt:
                break
            self.command("PRIVMSG", channel, txt)

    def parsing(self, txt):
        """ parse a line of txt received from the server and turn it into a IRC Event. """
        rawstr = str(txt)
        obj = Event()
        obj.txt = ""
        obj.cc = self.cc
        obj.command = ""
        obj.arguments = []
        arguments = rawstr.split()
        obj.origin = arguments[0]
        if obj.origin.startswith(":"):
            obj.origin = obj.origin[1:]
            if len(arguments) > 1:
                obj.command = arguments[1]
            if len(arguments) > 2:
                txtlist = []
                adding = False
                for arg in arguments[2:]:
                    if arg.startswith(":"):
                        adding = True
                        txtlist.append(arg[1:])
                        continue
                    if adding:
                        txtlist.append(arg)
                    else:
                        obj.arguments.append(arg)
                    obj.txt = " ".join(txtlist)
        else:
            obj.cmd = obj.command = obj.origin
            obj.origin = self.cfg.server
        try:
            obj.nick, obj.origin = obj.origin.split("!")
        except:
            obj.nick = ""
        if obj.arguments:
            obj.target = obj.arguments[-1]
        else:
            obj.target = ""
        if obj.target.startswith("#"):
            obj.channel = obj.target
        else:
            obj.channel = obj.nick
        if not obj.txt:
            if rawstr[0] == ":":
                rawstr = rawstr[1:]
            obj.txt = rawstr.split(":", 1)[-1]
        if not obj.txt and len(arguments) == 1:
            obj.txt = arguments[1]
        obj.args = obj.txt.split()
        obj.cb = obj.command
        obj.orig = repr(self)
        obj.rest = " ".join(obj.args)
        return obj

    @locked
    def raw(self, txt, direct=False):
        """ locked function that does the output to the IRC server, uses a 3 seconds sleep between messages. """
        if self._stopped:
            return
        if not txt.startswith("PING") and not txt.startswith("PONG"):
            logging.warn("%s" % txt)
        if not txt.endswith("\r\n"):
            txt += "\r\n"
        txt = txt[:512]
        txt = bytes(txt, "utf-8")
        if not direct and (time.time() - self._last) < 3.0:
            time.sleep(1.0 * (self.nrsend % 10))
        self._last = time.time()
        self.nrsend += 1
        self.sock.send(txt)

    def ready(self):
        """ flag the bot as ready. """
        self._ready.set()

    def register(self, k, v):
        """ register a handler with the bot. """
        self.handlers[k] = v

    def reload(self):
        """ reload cfg. """
        m = Selector("server=%s" % self.cfg.server)
        self.cfg.last(m)

    def say(self, channel, txt):
        """ say a text on a channel in max. 500 chars parts. """
        wrapper = TextWrap()
        txt = str(txt)
        for line in txt.split("\n"):
            for t in wrapper.wrap(line):
                self._outqueue.put((channel, t))

    def some(self, ssl=False, encoding="utf-8"):
        """ read from socket and add to buffer. """
        if ssl:
            inbytes = self.sock.read()
        else:
            inbytes = self.sock.recv(512)
        txt = str(inbytes, encoding)
        if txt == "":
            raise ConnectionResetError
        txt = bytes(txt, "utf-8").decode("unicode_escape")
        self.lastline += txt
        splitted = self.lastline.split("\r\n")
        for s in splitted[:-1]:
            self.buffer.append(s)
            if not s.startswith("PING") and not s.startswith("PONG"):
                logging.warn("%s" % s.strip())
        self.lastline = splitted[-1]

    def start(self):
        """ starts the IRC bot after reading the config file from disk, updated with runtime CLI options. """
        s = cfg.get("server", None) or self.cfg.get("server", "localhost")
        m = Selector("server=%s" % s)
        self.cfg.last(m)
        self.cfg.upgrade(cfg)
        self.cfg.save()
        if self.cfg.channel:
            self.channels.append(self.cfg.channel)
        nr = 0
        self._error = ""
        while 1:
            try:
                self.connect()
                break
            except ConnectionRefusedError:
                logging.warn("%s connection refused" % self.cfg.server)
                time.sleep(nr * 3.0)
            nr += 1
        if not self._error:
            self.logon(self.cfg.username or self.cfg.nick, self.cfg.realname or self.cfg.nick)
            super().start()
            launch(self.output)
        return

    def stop(self):
        """ stop the bot. """
        logging.error("stop %s" % self.cfg.server)
        self._stopped = True
        super().stop()
        self._outqueue.put_nowait((None, None))
        self.ready()

    def wait(self):
        """ wait for the bot to be ready. """
        self._ready.wait()

## commands

def reconnect(event):
    """ reconnect the bot. """
    bot = fleet.get_bot(event.orig)
    if bot and type(bot) == obj.bots.irc.IRC:
        bot._sock.shutdown(2)

def op(event):
    """ give operator status. """
    bot = fleet.get_bot(event.orig)
    if bot and type(bot) == obj.bots.irc.IRC:
        bot.command("MODE", event.channel, "+o", event.nick)

